local NB = NotaBene
local L = NB.L
local LSM = LibStub( "LibSharedMedia-3.0" )

local GetSpellCooldown = GetSpellCooldown
local GetSpellInfo = GetSpellInfo

local Module = NotaBene:NewModule( "Cooldowns", "AceEvent-3.0", "AceTimer-3.0", "LibSink-2.0" )

local registry = {}
local namedSpellCache = {}
local identitySpellCache = {}
local soundSchemes =
{
	["None"] = {},
}

local soundTypes = {
	None = L["None"],
	LSM = L["Predefined sound"],
	Custom = L["Custom sound"],
	Scheme = L["Scheme-defined sound"],
}

local defaults = {
	profile = {
		durationThreshold = 5,
		updateInterval = 0.5,
		
		soundScheme = "None",

		sinkOptions = {
			sink20OutputSink = "Default",
		},
		
		defaultAnnounces = {
			ready = {
				showMessage = true,
				messageText = L["%S:ffff0000 is ready!"],
				useMessageColor = false,
				useMessageFont = false,
				useMessageSize = false,
				messageColor = { 0, 1, 0 },
				messageSize = 26,
				
				soundType = "Scheme",
			},
			prelude = {
				showMessage = true,
				messageText = L["%S:ffff0000 in %T seconds!"],
				useMessageColor = false,
				useMessageFont = false,
				useMessageSize = false,
				messageColor = { 0, 1, 0 },
				messageSize = 26,
				
				soundType = "Scheme",
			},
		},
			
		spells = {
			["**"] = {
				enabled = true,
				preludeTime = 0,

				announces = {
					ready = {
						useDefault = true,

						showMessage = true,
						messageText = L["%S:ffff0000 is ready!"],
						useMessageColor = false,
						useMessageFont = false,
						useMessageSize = false,
						messageColor = { 0, 1, 0 },
						messageSize = 26,
						
						soundType = "Scheme",
					},
					prelude = {
						useDefault = true,
						
						showMessage = true,
						messageText = L["%S:ffff0000 in %T seconds!"],
						useMessageColor = false,
						useMessageFont = false,
						useMessageSize = false,
						messageColor = { 0, 1, 0 },
						messageSize = 26,
						
						soundType = "Scheme",
					},
				},
			},
		},
	},
}

function Module:OnInitialize()
	self.db = NB.db:RegisterNamespace( "Cooldowns", defaults )

	self.delayed = {}
	self.active = {}
	
	self:SetupOptions()
	
	self:RegisterEvent( "UNIT_SPELLCAST_SUCCEEDED" )
	self:RegisterEvent( "SPELL_UPDATE_COOLDOWN" )
end

function Module:HandleProfileChanges( event )
	self:SetSinkStorage( self.db.profile.sinkOptions )
end

local function NormalizeSpellName( name, rank )
	if not rank or #rank == 0 then
		if name:find( "%(.+%)$" ) then
			name = name.."()"
		end
	else
		name = name.."("..rank..")"
	end
	
	return name
end

function Module:RegisterSpell( key, ... )
	local spellCount = select( "#", ... )
	
	assert( type( key ) == "string" )
	assert( spellCount > 0 )
	
	for i = 1, spellCount do
		local spellId = select( i, ... )
		
		assert( type( spellId ) == "number" )
		
		local name, rank = GetSpellInfo( spellId )
		
		if not name then
			error( "Attempt to cache invalid spell: "..spellId )
		end
		
		local spellInfo = {}
		
		spellInfo.id = spellId
		spellInfo.key = key
		spellInfo.name = name
		spellInfo.rank = rank
		
		if i == 1 then
			registry[key] = name
		end
		
		name = NormalizeSpellName( name, rank )
		
		spellInfo.fullName = name
		
		namedSpellCache[name] = spellInfo
		identitySpellCache[spellId] = spellInfo
	end
end

function Module:IterateSpells()
	return pairs( registry )
end

function Module:RegisterSoundScheme( name, scheme )
	soundSchemes[name] = scheme
end

function Module:UNIT_SPELLCAST_SUCCEEDED( event, unit, spellName, spellRank )
	if unit == "player" then
		spellName = NormalizeSpellName( spellName, spellRank )
		
		local spellInfo = namedSpellCache[spellName]
		
		if spellInfo then
			local spellSettings = self.db.profile.spells[spellInfo.key]
			
			if spellSettings.enabled then
				NB:Debug( "%s detected", spellName )
				self.lastSuccededSpell = spellInfo
			end
		end
	end
end

function Module:SPELL_UPDATE_COOLDOWN( event )
	local now = GetTime()
	
	for spellName in pairs( self.delayed ) do
		local start, duration, enabled = GetSpellCooldown( spellName )
		
		if enabled == 1 then
			if duration > self.db.profile.durationThreshold then
				local spellInfo = namedSpellCache[spellName]
				
				NB:Debug( "%s activated (%d)", spellName, duration )
				
				self.active[spellInfo.id] = {
					time = now + duration,
				}
			end

			self.delayed[spellName] = nil
		end
	end
	
	if self.lastSuccededSpell then
		local lss = self.lastSuccededSpell
		local start, duration, enabled = GetSpellCooldown( lss.fullName )
		
		if enabled == 1 then
			if duration > self.db.profile.durationThreshold then
				NB:Debug( "%s activated (%d)", lss.fullName, duration )
				self.active[lss.id] = {
					time = now + duration,
				}
			end
		else
			self.delayed[lss.fullName] = true
		end
		
		self.lastSuccededSpell = nil
	end
	
	if next( self.active ) and not self.timer then
		self.timer = self:ScheduleRepeatingTimer( "ProcessUpdate", self.db.profile.updateInterval )
	end
end

local preludeArgs = { S = { type = 'text', }, T = { type = 'time', }, }
local readyArgs = { S = { type = 'text', }, }

function Module:ProcessUpdate()
	local now = GetTime()
	
	for spellId, data in pairs( self.active ) do
		local spellInfo = identitySpellCache[spellId]
		local spellSettings = self.db.profile.spells[spellInfo.key]
		local remaining = data.time - now

		if remaining <= 0 then
			readyArgs.S.value = spellInfo.name
			
			self:Announce( spellInfo.key, "ready", spellSettings, readyArgs )
			
			self.active[spellId] = nil
		elseif spellSettings.preludeTime > 0 and not data.prelude and remaining < spellSettings.preludeTime then
			preludeArgs.S.value = spellInfo.name
			preludeArgs.T.value = spellSettings.preludeTime
			
			self:Announce( spellInfo.key, "prelude", spellSettings, preludeArgs )
			
			data.prelude = true
		end
	end
	
	if not next( self.active ) then
		self:CancelTimer( self.timer, true )
		self.timer = nil
	end
end

function Module:GetSoundSchemes()
	return soundSchemes
end

function Module:Announce( key, event, spellSettings, formatArgs )
	local announce = spellSettings.announces[event]

	if announce.useDefault then
		announce = self.db.profile.defaultAnnounces[event]
	end
	
	if announce.showMessage then
		local message = announce.messageText
		
		if message then
			local r, g, b, font, size
			
			if announce.useMessageColor and announce.messageColor then
				local color = announce.messageColor
				
				if color then
					r, g, b = unpack( color )
				end
			end
			
			if announce.useMessageFont and announce.messageFont then
				font = announce.messageFont --LSM:Fetch( LSM.MediaType.FONT, announce.messageFont, true )
			end
			
			if announce.useMessageSize and announce.messageSize then
				size = announce.messageSize
			end
			
			message = NB:FormatMessage( message, formatArgs )
			
			self:Pour( message, r, g, b, font, size )
		end
	end
	
	local soundFile
	
	if announce.soundType == "LSM" and announce.soundLSM then
		soundFile = LSM:Fetch( LSM.MediaType.SOUND, announce.soundLSM, true )
	elseif announce.soundType == "Custom" and type( announce.soundFile ) == "string" and #announce.soundFile > 0 then
		soundFile = announce.soundFile
	elseif announce.soundType == "Scheme" then
		local scheme = soundSchemes[self.db.profile.soundScheme]
		
		if scheme then
			soundFile = scheme[key.."_"..event]
			
			if not soundFile then
				soundFile = scheme["Default_"..event]
			end
		end
	end
	
	if soundFile then
		PlaySoundFile( soundFile )
	end
end

function Module:BuildAnnounceOptions( options, dbPath, prefix, useDefault )
	local isReady = prefix == "ready"
	
	local group = {
		type = "group",
		order = isReady and -2 or -1,
		name = isReady and L["Cooldown ready announce"] or L["Cooldown prelude announce"],
		childGroups = "tab",
		get = function( info ) return dbPath[prefix][info.arg or info[#info]] end,
		set = function( info, value ) dbPath[prefix][info.arg or info[#info]] = value end,
		args = {
			message = {
				type = "group",
				name = L["Message"],
				args = {
					showMessage = {
						type = "toggle",
						name = L["Show message"],
						width = "full",
						order = 41,
					},
					messageText = {
						type = "input",
						name = L["Message"],
						width = "full",
						order = 42,
						disabled = function() return not dbPath[prefix].showMessage end,
					},
					useMessageColor = {
						type = "toggle",
						name = L["Override message color"],
						order = 43,
						disabled = function() return not dbPath[prefix].showMessage end,
					},
					messageColor = {
						type = "color",
						name = L["Message color"],
						get = function( info ) return unpack( dbPath[prefix][info[#info]] ) end,
						set = function( info, r, g, b ) dbPath[prefix][info[#info]] = { r, g, b } end,
						order = 44,
						disabled = function() return not( dbPath[prefix].showMessage ) or not( dbPath[prefix].useMessageColor ) end,
					},
					useMessageFont = {
						type = "toggle",
						name = L["Override message font"],
						order = 45,
						disabled = function() return not dbPath[prefix].showMessage end,
					},
					messageFont = {
						name = L["Font"],
						type = "select",
						order = 46,
						dialogControl = "LSM30_Font",
						values = function() return AceGUIWidgetLSMlists.font end,
						disabled = function() return not( dbPath[prefix].showMessage ) or not( dbPath[prefix].useMessageFont ) end,
					},
					useMessageSize = {
						type = "toggle",
						name = L["Override message size"],
						order = 47,
						disabled = function() return not dbPath[prefix].showMessage end,
					},
					messageSize = {
						name = L["Font size"],
						type = "range",
						order = 48,
						min = 4,
						max = 38,
						step = 1,
						bigStep = 1,
						disabled = function() return not( dbPath[prefix].showMessage ) or not( dbPath[prefix].useMessageSize ) end,
					},
				},
			},
			sound = {
				type = "group",
				name = L["Sound"],
				args = {
					soundType = {
						type = "select",
						name = L["Sound type"],
						width = "full",
						order = 51,
						values = soundTypes,
					},
					soundLSM = {
						name = L["Predefined sound"],
						type = "select",
						order = 52,
						width = "full",
						dialogControl = "LSM30_Sound",
						values = function() return AceGUIWidgetLSMlists.sound end,
						hidden = function() return dbPath[prefix].soundType ~= "LSM" end,
						disabled = function() return dbPath[prefix].soundType ~= "LSM" end,
					},
					soundFile = {
						type = "input",
						name = L["Sound file"],
						width = "full",
						order = 53,
						hidden = function() return dbPath[prefix].soundType ~= "Custom" end,
						disabled = function() return dbPath[prefix].soundType ~= "Custom" end,
					},
				},
			},
		},
	}
	
	if useDefault then
		group.args.useDefault = {
			type = "toggle",
			name = L["Use defaults"],
			order = 1,
			disabled = false,
		}
		
		group.args.copyDefaults = {
			type = "execute",
			name = L["Copy from defaults"],
			order = 2,
			disabled = false,
			func = function()
				NB:CopyTable( self.db.profile.defaultAnnounces[prefix], dbPath[prefix] )
				dbPath[prefix].useDefault = false
			end,
		}
		
		local disabled = function() return dbPath[prefix].useDefault end
		
		for k, v in pairs( group.args ) do
			if v.disabled == nil then
				v.disabled = disabled
			end
		end
	end
	
	options[prefix] = group
	
	return group
end

function Module:SetupSpellOptions( key )
	local spellInfo = namedSpellCache[registry[key]]
	
	local spell = {
		type = "group",
		name = registry[key],
		childGroups = "tab",
		args = {
			enabled = {
				type = "toggle",
				name = L["Enable"],
				order = 1,
				get = function() return self.db.profile.spells[key].enabled end,
				set = function( info, value ) self.db.profile.spells[key].enabled = value end,
			},
			preludeTime = {
				type = "range",
				name = L["Prelude time"],
				min = 0,
				max = 60,
				step = 1,
				order = 2,
				get = function() return self.db.profile.spells[key].preludeTime end,
				set = function( info, value ) self.db.profile.spells[key].preludeTime = value end,
			},
		},
	}
	
	self:BuildAnnounceOptions( spell.args, self.db.profile.spells[key].announces, "ready", true )
	self:BuildAnnounceOptions( spell.args, self.db.profile.spells[key].announces, "prelude", true )
	
	return spell
end

local options = {
	type = "group",
	name = L["Cooldowns"],
	args = {
		durationThreshold = {
			type = "range",
			min = 3,
			max = 15,
			step = 1,
			order = 1,
			name = L["Minimum cooldown duration"],
			get = function() return Module.db.profile.durationThreshold end,
			set = function( info, value ) Module.db.profile.durationThreshold = value end,
		},
		updateInterval = {
			type = "range",
			min = 0.1,
			max = 2.0,
			step = 0.1,
			order = 2,
			name = L["Update interval"],
			get = function() return Module.db.profile.updateInterval end,
			set = function( info, value ) Module.db.profile.updateInterval = value end,
		},
		soundScheme = {
			type = "select",
			width = "full",
			order = 3,
			name = L["Sound Scheme"],
			desc = L["Sound scheme to use"],
			values = function()
				local list = {}
				
				for k in pairs( soundSchemes ) do
					list[k] = k
				end
				
				return list
			end,
			get = function() return Module.db.profile.soundScheme end,
			set = function( info, value )
				Module.db.profile.soundScheme = value
			end,
		},
		sink = Module:GetSinkAce3OptionsDataTable(),
		default = {
			type = "group",
			childGroups = "tab",
			order = 5,
			name = "|cffffff00"..L["Default settings"].."|r",
			args = {
			},
		},
	},
}

options.args.sink.order = 4
options.args.sink.inline = true

function Module:SetupOptions()
	self:BuildAnnounceOptions( options.args.default.args, self.db.profile.defaultAnnounces, "ready" )
	self:BuildAnnounceOptions( options.args.default.args, self.db.profile.defaultAnnounces, "prelude" )
	
	for k in pairs( registry ) do
		options.args[k] = self:SetupSpellOptions( k )
	end
	
	NB:AddToplevelOptionsSection( "cooldowns", options )
end