local NB = NotaBene
local L = NB.L
local LSM = LibStub( "LibSharedMedia-3.0" )

local GetSpellInfo = GetSpellInfo

local Module = NotaBene:NewModule( "Procs", "AceEvent-3.0", "LibSink-2.0" )

local registry = {}
local identitySpellCache = {}
local soundSchemes =
{
	["None"] = {},
}

local soundTypes = {
	None = L["None"],
	LSM = L["Predefined sound"],
	Custom = L["Custom sound"],
	Scheme = L["Scheme-defined sound"],
}

local defaults = {
	profile = {
		soundScheme = "None",
		
		defaultAnnounce = {
			showMessage = true,
			messageText = L["%S!"],
			useMessageColor = false,
			useMessageFont = false,
			useMessageSize = false,
			messageColor = { 0, 1, 0 },
			messageSize = 26,
			
			soundType = "Scheme",
		},

		sinkOptions = {
			sink20OutputSink = "Default",
		},

		spells = {
			["**"] = {
				enabled = true,

				announce = {
					useDefault = true,

					showMessage = true,
					messageText = L["%S!"],
					useMessageColor = false,
					useMessageFont = false,
					useMessageSize = false,
					messageColor = { 0, 1, 0 },
					messageSize = 26,
					
					soundType = "Scheme",
				},
			},
		},
	},
}

function Module:OnInitialize()
	self.db = NB.db:RegisterNamespace( "Procs", defaults )

	self.playerName = UnitName( "player" )
	
	self:SetupOptions()
	
	self:RegisterEvent( "COMBAT_LOG_EVENT_UNFILTERED" )
end

function Module:HandleProfileChanges( event )
	self:SetSinkStorage( self.db.profile.sinkOptions )
end

function Module:RegisterSpell( key, spellId, selfOnly, ... )
	local typeCount = select( "#", ... )
	
	assert( type( key ) == "string" )
	assert( type( spellId ) == "number" or type( spellId ) == "table" )
	assert( typeCount > 0 )
	
	if type( spellId ) == "number" then
		spellId = { spellId }
	end
	
	local name = GetSpellInfo( spellId[1] )
		
	if not name then
		error( "Attempt to cache invalid spell: "..spellId[1] )
	end
	
	for i, sid in ipairs( spellId ) do
		local spellInfo = {}
	
		spellInfo.id = sid
		spellInfo.key = key
		spellInfo.name = name
		spellInfo.types = { ... }
		
		if i == 1 then
			registry[key] = spellInfo
		end
		
		identitySpellCache[sid] = spellInfo
	end
end

function Module:IterateSpells()
	return pairs( registry )
end

function Module:RegisterSoundScheme( name, scheme )
	soundSchemes[name] = scheme
end

local function contains( item, list )
	for k, v in ipairs( list ) do
		if v == item then
			return true
		end
	end
end

function Module:COMBAT_LOG_EVENT_UNFILTERED( event, ... )
	local msg, type, msg2, name, msg4, msg4, name2 = select( 1, ... )
	local spellId, spellName, spellSchool = select( 9, ... )
	
	if identitySpellCache[spellId] then
		local spellInfo = identitySpellCache[spellId]
		local spellSettings = self.db.profile.spells[spellInfo.key]
		
		if spellSettings.enabled and contains( type, spellInfo.types ) then
			if name == self.playerName then
				if spellInfo.selfOnly and name2 == self.playerName then
					self:Announce( spellInfo )
				elseif not( spellInfo.selfOnly ) then
					self:Announce( spellInfo )
				end
			elseif name == nil and name2 == self.playerName then
				self:Announce( spellInfo )
			end
		end
	end
end

function Module:GetSoundSchemes()
	return soundSchemes
end

local procArgs = { S = { type = 'text', }, }

function Module:Announce( spellInfo )
	local announce = self.db.profile.spells[spellInfo.key].announce

	if announce.useDefault then
		announce = self.db.profile.defaultAnnounce
	end
	
	if announce.showMessage then
		local message = announce.messageText
		
		if message then
			local r, g, b, font, size
			
			if announce.useMessageColor and announce.messageColor then
				local color = announce.messageColor
				
				if color then
					r, g, b = unpack( color )
				end
			end
			
			if announce.useMessageFont and announce.messageFont then
				font = announce.messageFont --LSM:Fetch( LSM.MediaType.FONT, announce.messageFont, true )
			end
			
			if announce.useMessageSize and announce.messageSize then
				size = announce.messageSize
			end
			
			procArgs.S.value = spellInfo.name
			message = NB:FormatMessage( message, procArgs )
			
			--print( ("%s: (%d, %d, %d) / %s / %d"):format( spellInfo.key, r or -1, g or -1, b or -1, font or "<none>", size or -1 ) )
			self:Pour( message, r, g, b, font, size )
		end
	end
	
	local soundFile
	
	if announce.soundType == "LSM" and announce.soundLSM then
		soundFile = LSM:Fetch( LSM.MediaType.SOUND, announce.soundLSM, true )
	elseif announce.soundType == "Custom" and type( announce.soundFile ) == "string" and #announce.soundFile > 0 then
		soundFile = announce.soundFile
	elseif announce.soundType == "Scheme" then
		local scheme = soundSchemes[self.db.profile.soundScheme]
		
		if scheme then
			soundFile = scheme[spellInfo.key]
			
			if not soundFile then
				soundFile = scheme["Default"]
			end
		end
	end
	
	if soundFile then
		PlaySoundFile( soundFile )
	end
end

function Module:BuildAnnounceOptions( options, dbPath, useDefault )
	local group = {
		type = "group",
		name = L["Announce"],
		childGroups = "tab",
		get = function( info ) return dbPath[info.arg or info[#info]] end,
		set = function( info, value ) dbPath[info.arg or info[#info]] = value end,
		args = {
			message = {
				type = "group",
				name = L["Message"],
				args = {
					showMessage = {
						type = "toggle",
						name = L["Show message"],
						width = "full",
						order = 41,
					},
					messageText = {
						type = "input",
						name = L["Message"],
						width = "full",
						order = 42,
						disabled = function() return not dbPath.showMessage end,
					},
					useMessageColor = {
						type = "toggle",
						name = L["Override message color"],
						order = 43,
						disabled = function() return not dbPath.showMessage end,
					},
					messageColor = {
						type = "color",
						name = L["Message color"],
						get = function( info ) return unpack( dbPath[info[#info]] ) end,
						set = function( info, r, g, b ) dbPath[info[#info]] = { r, g, b } end,
						order = 44,
						disabled = function() return not( dbPath.showMessage ) or not( dbPath.useMessageColor ) end,
					},
					useMessageFont = {
						type = "toggle",
						name = L["Override message font"],
						order = 45,
						disabled = function() return not dbPath.showMessage end,
					},
					messageFont = {
						name = L["Font"],
						type = "select",
						order = 46,
						dialogControl = "LSM30_Font",
						values = function() return AceGUIWidgetLSMlists.font end,
						disabled = function() return not( dbPath.showMessage ) or not( dbPath.useMessageFont ) end,
					},
					useMessageSize = {
						type = "toggle",
						name = L["Override message size"],
						order = 47,
						disabled = function() return not dbPath.showMessage end,
					},
					messageSize = {
						name = L["Font size"],
						type = "range",
						order = 48,
						min = 4,
						max = 38,
						step = 1,
						bigStep = 1,
						disabled = function() return not( dbPath.showMessage ) or not( dbPath.useMessageSize ) end,
					},
				},
			},
			sound = {
				type = "group",
				name = L["Sound"],
				args = {
					soundType = {
						type = "select",
						name = L["Sound type"],
						width = "full",
						order = 51,
						values = soundTypes,
					},
					soundLSM = {
						name = L["Predefined sound"],
						type = "select",
						order = 52,
						width = "full",
						dialogControl = "LSM30_Sound",
						values = function() return AceGUIWidgetLSMlists.sound end,
						hidden = function() return dbPath.soundType ~= "LSM" end,
						disabled = function() return dbPath.soundType ~= "LSM" end,
					},
					soundFile = {
						type = "input",
						name = L["Sound file"],
						width = "full",
						order = 53,
						hidden = function() return dbPath.soundType ~= "Custom" end,
						disabled = function() return dbPath.soundType ~= "Custom" end,
					},
				},
			},
		},
	}
	
	if useDefault then
		group.args.useDefault = {
			type = "toggle",
			name = L["Use defaults"],
			order = 1,
			disabled = false,
		}
		
		group.args.copyDefaults = {
			type = "execute",
			name = L["Copy from defaults"],
			order = 2,
			disabled = false,
			func = function()
				NB:CopyTable( self.db.profile.defaultAnnounce, dbPath )
				dbPath.useDefault = false
			end,
		}
		
		local disabled = function() return dbPath.useDefault end
		
		for k, v in pairs( group.args ) do
			if v.disabled == nil then
				v.disabled = disabled
			end
		end
	end
	
	options.announce = group
	
	return group
end

function Module:SetupSpellOptions( key )
	local spellInfo = registry[key]
	
	local spell = {
		type = "group",
		name = spellInfo.name,
		childGroups = "tab",
		args = {
			enabled = {
				type = "toggle",
				name = L["Enable"],
				order = 1,
				get = function() return self.db.profile.spells[key].enabled end,
				set = function( info, value ) self.db.profile.spells[key].enabled = value end,
			},
		},
	}
	
	self:BuildAnnounceOptions( spell.args, self.db.profile.spells[key].announce, true )
	
	return spell
end

local options = {
	type = "group",
	name = L["Procs"],
	args = {
		soundScheme = {
			type = "select",
			width = "full",
			order = 1,
			name = L["Sound Scheme"],
			desc = L["Sound scheme to use"],
			values = function()
				local list = {}
				
				for k in pairs( soundSchemes ) do
					list[k] = k
				end
				
				return list
			end,
			get = function() return Module.db.profile.soundScheme end,
			set = function( info, value )
				Module.db.profile.soundScheme = value
			end,
		},
		sink = Module:GetSinkAce3OptionsDataTable(),
		default = {
			type = "group",
			childGroups = "tab",
			order = 3,
			name = "|cffffff00"..L["Default settings"].."|r",
			args = {
			},
		},
	},
}

options.args.sink.order = 2
options.args.sink.inline = true

function Module:SetupOptions()
	self:BuildAnnounceOptions( options.args.default.args, self.db.profile.defaultAnnounce )
	
	for k in pairs( registry ) do
		options.args[k] = self:SetupSpellOptions( k )
	end
	
	NB:AddToplevelOptionsSection( "procs", options )
end